-- Radial tiling

module Rings where

import Monad
import Pan
-- import CircleLimit

-- Make a given sized image into concentric rings of copies.  Each ring
-- gets as many copies as will fit, allowing at least a given separation.
rings :: FloatE -> SizedImageC -> ImageC
rings sep (w,h,im) p = im' p
 where
   im'        = ifE (imPerRing <=* 0) empty $
                rotate iAng $ translate (iDist,0) $ im
   (r,a)      = toPolar p
   iAng       = a - ifE (imPerRing <=* 0) 0 (wrap' angPerIm a)
   iDist      = ringDist -- add w/2 if ringDist uses wrap
   ringDist   = r - wrap' (w+sep) r       -- to ring center (base if wrap)
   ringLength = 2*pi * ringDist          -- circumf of ring base
   imPerRing  = floorE (ringLength/(h+sep)) -- # ims per ring
   angPerIm   = 2*pi / i2f imPerRing

counterTurnP :: BoolE -> FloatE -> FloatE -> TransformE

counterTurnP flipDirs ringWidth ang p = xf p
 where
   dir     = ifE (uscale ringWidth rings p) (-1) 1
   rings p = evenE (floorE (distO p - 0.5))
   xf = ifE flipDirs (rotateP ang . scaleP (dir,1))
                     (rotateP (dir * ang))

{-
counterTurnP flipDirs ringWidth ang p = xf p
 where
   dir            = ifE (oddE ringNum) (-1) 1
   ringNum        = floorE (distO p / ringWidth - 0.5)
   xf | flipDirs  = rotateP ang . scaleP (dir,1)
      | otherwise = rotateP (dir * ang)
-}

{-  Alternatively (but without flipDirs):

-- Generalization of rotate.
rotateI :: Image FloatE -> Filter c
rotateI angI im p = im (rotateP (- (angI p)) p)

counterTurnP ringWidth ang =
  rotateI (uscale ringWidth (cond rings (const ang) (const (-ang))))

rings = evenE . floorE . distO
-}


counterTurn :: BoolE -> FloatE -> FloatE -> Filter a
counterTurn flip rw ang = (. counterTurnP flip rw (-ang))

counterTurnUI :: UI FilterBC
counterTurnUI = do wid <- slider "ring width" (0,50) 10
                   return $ flip (counterTurn false wid)

counterTurnOscUI :: UI FilterBC
counterTurnOscUI = do filterB <- counterTurnUI
                      return $ \ im t -> filterB im (pi/2 * (1 + sin t))



progressFilterB :: UI FilterBC -> Double -> UI FilterC
progressFilterB fui maxT = 
  do filterB <- fui
     t       <- liftM (* fromDoub maxT) progressUI
     return $ \ im -> filterB im t
 where
   fromDoub = fromRational . toRational  -- fromDouble is Hugs only

-- Note that angPerIm will be infinite if imPerRing is zero, which happens
-- near the origin.

ringsUI = liftM2 rings
            (slider "separation" (-30,150) 0)
            (browseImageSize "image" "flower.bmp")

ringsCounterTurningUI =
  liftM3 ringsf
    (tick "Flip alternate rings" True)
    (slider "separation" (-30,150) 0)
    (browseImageSize "image" "flower.bmp")
 where
    ringsf flipDirs sep szim@(w,_,_) t =
      counterTurn flipDirs (w+sep) t (rings sep szim)

--------------------------------------------------

makeAll = makeAllFor allBackends

makeAllFor bends = do
  compileImageFor bends "ring tiling" ringsUI
  compileFilterBFor bends "counter-turning rings self-timed" $
    atMouseB counterTurnUI
  compileFilterBFor bends "counter-turning rings oscillating" $
    atMouseB counterTurnOscUI
  compileFilterFor bends "counter-turning rings" $
    atMouse (progressFilterB counterTurnUI (2*pi))
  compileImageBFor bends "ring tiling counter-turning self-timed"
    ringsCounterTurningUI
  compileImageFor   bends "ring tiling counter-turning" $
    ringsCounterTurningUI `ap'` progressAngleUI
