
#include "stdafx.h"
#include "Box.h"


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
CBox::CBox() 
{
      m_initialized = false;
      m_shuttingDown = false;
      m_isVertical = true;
      m_hGap = 10;
      m_wGap = 10;
      m_children = new std::vector<CWindowVirt*>();
}

CBox::CBox(bool isVertical)
{
    m_initialized = false;
    m_shuttingDown = false;
    m_isVertical = isVertical;
    m_hGap = 10;
    m_wGap = 10;
    m_children = new std::vector<CWindowVirt*>();
}

CBox::~CBox()
{
    HWND hWndChild;
    while (m_children->size() > 0) {
        hWndChild = (HWND)m_children->front();
        RemoveProps(hWndChild);
        delete m_children->front();
        m_children->erase(m_children->begin());
    }
    if (m_hWnd) {
        RemoveProps(m_hWnd);
        ::DestroyWindow(m_hWnd);
    }
    delete m_children;
}

//////////////////////////////////////////////////////////////////////
// Methods
//////////////////////////////////////////////////////////////////////

void
CBox::AddChild(CWindowVirt* ch)
{
    m_children->push_back(ch);
}

HWND
CBox::Create(LPCTSTR szLabel, int width, int height, HWND hWndParent)
{
      rc.left = 0;
      rc.right = width;
      rc.top = 0;
      rc.bottom = height;

      return CWindowImpl<CBox,CWindowVirt, BoxTraits>::Create(hWndParent, rc, szLabel);
}


//////////////////////////////////////////////////////////////////////
// Event handlers
//////////////////////////////////////////////////////////////////////

//
// FUNCTION: OnBgErase
//
// PURPOSE:  Handle WM_ERASEBKGND for a button groupbox in case the parent
//           has got WS_CLIPCHILDREN set.
//
// COMMENTS: see Q79982 for info on why this is needed.
//
LRESULT
CBox::OnBgErase(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
{
    HBRUSH  hBrush;
    HDC     hDC;
    RECT    rect;

    hDC = GetDC();

    // Obtain a handle to the parent window's background brush.
    hBrush = (HBRUSH)(COLOR_WINDOW+1);

    // Erase the group box's background.
    GetClientRect(&rect);
    FillRect(hDC, &rect, (HBRUSH)(COLOR_WINDOW));

    ReleaseDC(hDC);
//    DeleteObject(hDC);

    // Instruct Windows to paint the group box text and frame.
    InvalidateRect(NULL, FALSE);

    // Insert code here to instruct the contents of the group box
    // to repaint as well.

    return TRUE; // Background has been erased.
}

LRESULT
CBox::OnDestroy(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
{
    HWND hWndChild;

    hWndChild = GetWindow(GW_CHILD);
    while (hWndChild) {
        RemoveProps(hWndChild);
        hWndChild = ::GetNextWindow(hWndChild, GW_HWNDNEXT);
    }

    if (m_hWnd) RemoveProps(m_hWnd);

    bHandled = FALSE;
    return 0;
}

LRESULT
CBox::OnPropagateCmd(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
{
      // 
      // We pass along WM_COMMANDs up to the parent, so that
      // the user can handle these at the WndProc level where
      // the layout control is inserted.
      //
      // Ditto for WM_NOTIFYs.
      // 
      HWND hWndParent = GetParent();
      
      return SendMessage(hWndParent, uMsg, wParam, lParam);
}

LRESULT
CBox::OnSize(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
{
    int cx, cy;

    cx = LOWORD(lParam);
    cy = HIWORD(lParam);

    if (m_initialized && !m_shuttingDown) {
        if (m_isVertical) {
            VResize(cx,cy);
        } else {
            HResize(cx,cy);
        }
    }
    bHandled = FALSE;
    return 0;
}

//
// FUNCTION:    RemoveProps
//
// PURPOSE:     Remove the window properties that the box
//              added to all the child windows.
//
void
CBox::RemoveProps(HWND hWnd)
{
    // let go of the per-window properties.
    void* origSize = (void*)GetProp(hWnd, RESIZE_PROPERTY_ORIGSIZE);

    if (origSize) {
        free(origSize);
    }
    RemoveProp(hWnd, RESIZE_PROPERTY_ORIGSIZE);
    RemoveProp(hWnd, RESIZE_PROPERTY_HORIZ);
    RemoveProp(hWnd, RESIZE_PROPERTY_VERT);

    m_shuttingDown = true;

}

SIZE CBox::HResize(int width, int height)
{
    RECT r;
    int x, y;
    HWND hWndChild;
    SIZE sz;
    SIZE* pOrigSize;
    SIZE* pSize;
  
    int newWidth, newHeight, rszWidth;
    int theWidth, theHeight, prop;
  
    x = m_wGap;
    y = 2*m_hGap;
  
    if (height < 0) { 
        // Compute natural size of hbox:

        newWidth  = x;
        newHeight = 0;
    
        hWndChild = GetWindow(GW_CHILD);
        while (hWndChild) {

          if ( (pSize = (SIZE*)GetProp(hWndChild, RESIZE_PROPERTY_ORIGSIZE)) != NULL ) {
	        newWidth  += pSize->cx + m_wGap;
	        newHeight =  max (pSize->cy, newHeight);
          } else {
              ::GetWindowRect(hWndChild, &r);
	          newWidth  += (r.right - r.left) + m_wGap;
	          newHeight  = max(r.bottom - r.top, newHeight);
          }
          hWndChild  = ::GetNextWindow(hWndChild, GW_HWNDNEXT);
        }
    
        // add the top and bottom bands
        newHeight += y + m_hGap;
        // Allocate property holding original box size.
        if ( (pOrigSize = (SIZE*)GetProp(m_hWnd, RESIZE_PROPERTY_ORIGSIZE)) == NULL ) {
            pOrigSize = (SIZE*)malloc(sizeof(SIZE));

            if (pOrigSize) {
	            pOrigSize->cx = newWidth;
	            pOrigSize->cy = newHeight;
	            SetProp(m_hWnd, RESIZE_PROPERTY_ORIGSIZE, (HANDLE)pOrigSize);
            }
        }
        rszWidth = newWidth;
    } else {
        //
        // Resizing the hbox:
        //
        //   Figure out how much space needs to be allocated
        //   to the horizontally non-resizable controls =>
        //   how much is left
        //
      
        newHeight = height - 2*m_hGap; //y - m_hGap;
      
        pOrigSize = (SIZE*)GetProp(m_hWnd, RESIZE_PROPERTY_ORIGSIZE);

        //
        // amount we have of natural size to spend -- nat. size minus the gap
        // on either side.
        //
        rszWidth = pOrigSize->cx - x;
        // size to fill up.
        newWidth = width - x;

        hWndChild = GetWindow(GW_CHILD);

        while (hWndChild) {
            if ((int)GetProp(hWndChild, RESIZE_PROPERTY_HORIZ) != RESIZE_NORMAL) {
	        // Not resizable, substract (natural) size of this child from
	        // the amount we have to fill up.
	            if ( (pSize = (SIZE*)GetProp(hWndChild, RESIZE_PROPERTY_ORIGSIZE)) != NULL ) {
	                newWidth -= pSize->cx + m_wGap;
	                rszWidth -= pSize->cx + m_wGap;
	            } else {
                    ::GetWindowRect(hWndChild, &r);
	                newWidth -= (r.right - r.left) + m_wGap;
	                rszWidth -= (r.right - r.left) + m_wGap;
	            }
            } else {
	            newWidth -= m_wGap;
	            rszWidth -= m_wGap;
            }
            hWndChild = ::GetNextWindow(hWndChild, GW_HWNDNEXT);
        }
    }
  
    // newWidth now holds the amount we have to allocate to the
    // resizable children, rszWidth the natural size of these components.
    // 
    // The gaps have also been accounted for.

    hWndChild = GetWindow(GW_CHILD);
    while (hWndChild) {
        pSize = (SIZE*)GetProp(hWndChild, RESIZE_PROPERTY_ORIGSIZE);
        ::GetWindowRect(hWndChild, &r);

        prop  = (int)GetProp(hWndChild, RESIZE_PROPERTY_VERT);
        if (prop == RESIZE_NONE) {
            theHeight = r.bottom - r.top;
        } else {
            theHeight = newHeight;
        }

        prop = (int)GetProp(hWndChild, RESIZE_PROPERTY_HORIZ);
        if (prop == RESIZE_NORMAL && width > 0) {
            int childWidth;

            if (pSize) {
	            childWidth = pSize->cx;
            } else {
	            childWidth = r.right - r.left;
            }
            //char msg[255];
            //sprintf(msg, "%d %d %d %d %d", newWidth, childWidth, rszWidth, pOrigSize->cx, w);
            //::MessageBox(NULL, msg, "hResizer", MB_OK);
            theWidth = (int)((float)newWidth * ((float) childWidth / rszWidth));
        } else {
            if (pSize) {
	           theWidth = pSize->cx;
            } else {
	           theWidth = r.right - r.left;
            }
        }
      
        ::MoveWindow(hWndChild,x,y, theWidth, theHeight, TRUE);
      
        newHeight  = max(theHeight, newHeight);
        x += theWidth + m_wGap;

        hWndChild = ::GetNextWindow(hWndChild, GW_HWNDNEXT);
    }

//    newHeight += m_hGap;
    newWidth = x;
    
    sz.cx = max (width, newWidth);
    sz.cy = max (height, newHeight);
    return sz;
}

// 
// FUNCTION: vResizer
// 
// PURPOSE: Lay out the children of a HWND vertically. 
//
// COMMENTS: When called with either width or height < 0, the
//           routine computes the natural size of the box &
//           updates the window property holding that info.
SIZE CBox::VResize(int width, int height)
{
    RECT r;
    int x, y;
    HWND hWndChild;
    SIZE sz;
    SIZE* pOrigSize;
    SIZE* pSize;
  
    int newWidth, newHeight, rszHeight;
    int theWidth, theHeight, prop;

    x = m_wGap;
    y = 2*m_hGap;
  
    if (width < 0) { 
      // Compute natural size of box.

      newWidth  = 0;
      newHeight = y;

      hWndChild = GetWindow(GW_CHILD);
      while (hWndChild) {
	      if ( (pSize = (SIZE*)GetProp(hWndChild, RESIZE_PROPERTY_ORIGSIZE)) != NULL ) {
	         newHeight += pSize->cy + m_hGap;
	         newWidth   =  max (pSize->cx, newWidth);
	      } else {
            ::GetWindowRect(hWndChild, &r);
	         newHeight += (r.bottom - r.top) + m_hGap;
	         newWidth  = max(r.right - r.left, newWidth);
	      }
         hWndChild = ::GetNextWindow(hWndChild, GW_HWNDNEXT);
      }

      // add the borders
      newWidth += x + m_wGap;

      // Done computing the natural size, record it by allocating
      // a property that will hold it.
      if ( (pOrigSize = (SIZE*)GetProp(m_hWnd, RESIZE_PROPERTY_ORIGSIZE)) == NULL ) {
	      pOrigSize = (SIZE*)malloc(sizeof(SIZE));
	    
	      if (pOrigSize) {
	         pOrigSize->cx = newWidth;
	         pOrigSize->cy = newHeight;
	         SetProp(m_hWnd, RESIZE_PROPERTY_ORIGSIZE, (HANDLE)pOrigSize);
	      }
      }
      rszHeight = newHeight;
    } else {
      
      // max width of a child is minus the borders of the box.
      newWidth = width-2*m_wGap;

      pOrigSize = (SIZE*)GetProp(m_hWnd, RESIZE_PROPERTY_ORIGSIZE);

      // rszHeight holds the natural height of the vertically resizable
      // children.
      //
      // newHeight holds the height that needs to be distributed between
      // the resizable children.
      //
      // Substract away the upper border for both of them.
      rszHeight = pOrigSize->cy - y;
      newHeight = height - y;

      // 
      // Subtract away the borders *and* the height of the children
      // that are marked as not being resizeable vertically.
      // 
      hWndChild = GetWindow(GW_CHILD);
      while (hWndChild) {
	      if ((int)GetProp(hWndChild, RESIZE_PROPERTY_VERT) != RESIZE_NORMAL) {
	         if ( (pSize = (SIZE*)GetProp(hWndChild, RESIZE_PROPERTY_ORIGSIZE)) != NULL ) {
	            newHeight -= pSize->cy + m_hGap;
	            rszHeight -= pSize->cy + m_hGap;
	         } else {
               ::GetWindowRect(hWndChild, &r);
	            newHeight -= (r.bottom - r.top) + m_hGap;
	            rszHeight -= (r.bottom - r.top) + m_hGap;
	         }
	      } else {
	         newHeight -= m_hGap;
	         rszHeight -= m_hGap;
	      }
         hWndChild = ::GetNextWindow(hWndChild, GW_HWNDNEXT);
      }
    }

    //
    // Having computed the sum of the natural heights of the
    // children that should be resized *and* the amount of 
    // height that should be distributed between them.
    //
    hWndChild = GetWindow(GW_CHILD);
    while (hWndChild) {
      pSize = (SIZE*)GetProp(hWndChild, RESIZE_PROPERTY_ORIGSIZE);
      ::GetWindowRect(hWndChild, &r);

      // Horizontally, a component is either stretched to fit 
      // the box or not. 
      prop = (int)GetProp(hWndChild, RESIZE_PROPERTY_HORIZ);
      if (prop == RESIZE_NONE) {
	      theWidth = r.right - r.left;
      } else {
	      theWidth = newWidth;
      }
	 
      prop = (int)GetProp(hWndChild, RESIZE_PROPERTY_VERT);
      if (prop == RESIZE_NORMAL && height > 0 ) {
	      int childHeight;

	      if (pSize) {
	         childHeight = pSize->cy;
	      } else {
	         childHeight = r.bottom - r.top;
	      }

	      // calculate the portion of the available height to
	      // allocate to child using the natural height of the 
	      // child and box.
	      theHeight = (int)((float)newHeight * ((float) childHeight / rszHeight));
      } else {
	      if (pSize) {
	         theHeight = pSize->cy;
	      } else {
	         theHeight = r.bottom - r.top;
	      }
      }

      ::MoveWindow(hWndChild,x,y, theWidth, theHeight, TRUE);

      newWidth = max(theWidth, newWidth);
      y += theHeight + m_hGap;

      hWndChild = ::GetNextWindow(hWndChild, GW_HWNDNEXT);
    }

    // ToDo: the resize functions play a dual role: 1) compute the
    // placement of the children given a size. 2) Compute initial size
    // of the box based on the preferred sizes of its children.

    // (conal) we were getting 2*m_wGap too much width.
    //newWidth += 2*m_wGap;
    newHeight = y;

    sz.cx = max (width, newWidth);
    sz.cy = max (height, newHeight);

    return sz;

}

//
// FUNCTION: RecalcSize(HWND)
//
// PURPOSE: After having added all the components to the layout
//          container, you need to *explicitly* call RecalcSize() 
//          in your code to have it adjust itself to a size big
//          enough to display all components (but no bigger.)


void CBox::CalcLayout()
{
   SIZE sz;
   int canResize = 0;
   HWND hWndChild;

   if (CBox::m_isVertical) {
      sz = VResize(-1,-1);
   } else {
      sz = HResize(-1,-1);
   }

   m_initialized = true;
		
   // Figure out the stretch properties of the box. The rule is:
   // if any of the components are stretchable in either X or Y, the box is too.
   //   
   hWndChild = GetWindow(GW_CHILD);
		
   while (hWndChild && !canResize) {
      canResize = ((int)GetProp(hWndChild, RESIZE_PROPERTY_HORIZ) == RESIZE_NORMAL);
      hWndChild = ::GetWindow(hWndChild, GW_HWNDNEXT);
   }

   if (canResize) {
      SetProp(m_hWnd, RESIZE_PROPERTY_HORIZ, (HANDLE)RESIZE_NORMAL);
   } else {
      SetProp(m_hWnd, RESIZE_PROPERTY_HORIZ, (HANDLE)RESIZE_NONE);
   }
   canResize = 0;

   hWndChild = GetWindow(GW_CHILD);
   while (hWndChild && !canResize) {
      canResize = ((int)GetProp(hWndChild, RESIZE_PROPERTY_VERT) == RESIZE_NORMAL);
      hWndChild = ::GetWindow(hWndChild, GW_HWNDNEXT);
   }

   if (canResize) {
      SetProp(m_hWnd, RESIZE_PROPERTY_VERT, (HANDLE)RESIZE_NORMAL);
   } else {
      SetProp(m_hWnd, RESIZE_PROPERTY_VERT, (HANDLE)RESIZE_NONE);
   }

   ::SetWindowPos(m_hWnd, HWND_TOP, 0,0,sz.cx, sz.cy, SWP_NOMOVE | SWP_NOZORDER);
   ::InvalidateRect(m_hWnd,NULL,FALSE);
}

