// PanTick.cpp: implementation of the CPanTick class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "PanTick.h"
#include "Box.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CPanTick::CPanTick(HWND          hWndParent,
                   LPCTSTR       szLabel,
                   int           nWidth,
                   int           nHeight,
                   int           initValue,
                   IntUpdater*   pUpd,
                   void*         pParam,
                   CInvalidator* pInvalid)
{
    SIZE* pSize;
    RECT r;

    m_updateValue = pUpd;
    m_pParam      = pParam;
    m_pInvalid    = pInvalid;

    if (initValue) m_isChecked = true;

    r.left = 0;
    r.top  = 0;
    r.right = nWidth;
    r.bottom = nHeight;

    // Combine object and window creation at the PanUI level.
    Create(hWndParent,r,szLabel);

    // make sure we've initialized correctly.
    initValue = m_updateValue->GetInt(m_pParam);

    pSize = (SIZE*)malloc(sizeof(SIZE));
    if (pSize) {
      GetWindowRect(&r);

      pSize->cx = r.right - r.left;
      pSize->cy = r.bottom - r.top;
      SetProp(m_hWnd, RESIZE_PROPERTY_ORIGSIZE, (HANDLE)pSize);
      SetProp(m_hWnd, RESIZE_PROPERTY_HORIZ, (HANDLE)RESIZE_NORMAL);
      SetProp(m_hWnd, RESIZE_PROPERTY_VERT,  (HANDLE)RESIZE_NONE);
    }
}

CPanTick::~CPanTick()
{

}


LRESULT
CPanTick::OnCheck(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
{
    // notify the 'listener' by side-effecting the shared state, followed
    // by invalidation.
    m_updateValue->UpdateInt(m_pParam, (BOOL)wParam == TRUE ? 1 : 0);
    m_pInvalid->Invalidate();

    // pass the event downwards by flagging it as not handled.
    bHandled = false;

    return 0;
}
